import inspect
import re
from typing import Any, Callable, Dict, Iterable, Optional, Type, get_origin

Recur = Callable[[Any, Type], Any]
Serializer = Callable[[Any, Type, Recur], Any]
Deserializer = Callable[[Any, Type, Recur], Any]


_SPLIT_CAMEL_CASE_PATTERN = re.compile(
    ".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


class SerdeError(Exception):
    """
    Signals an error during serialization or deserialization.
    """

    def __init__(self, user_message: str):
        self.user_message = user_message


def as_child(cls: Type) -> Type:
    """
    Marks the class to be serialized as one of its children. This will add an
    additional "type" field in the result, so the child can be deserialized
    properly.
    """
    assert inspect.isclass(cls), cls
    cls.__serde_serialize_as_child__ = cls  # type: ignore
    return cls


def should_serialize_as_child(cls: Type) -> bool:
    """
    Checks whether the given class should be serialized as a child.
    """
    assert inspect.isclass(cls), cls
    return hasattr(cls, "__serde_serialize_as_child__")


def all_lower_to_camel_case(name: str) -> str:
    """
    Converts a string from all_lower_case to camelCase.
    """
    assert name.islower(), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return parts[0] + "".join(p.capitalize() for p in parts[1:])


def all_upper_to_camel_case(name: str) -> str:
    """
    Converts a string from ALL_UPPER_CASE to camelCase.
    """
    assert name.isupper(), name

    if not name:
        return ""

    parts = name.split("_")
    assert parts, (name, parts)

    return parts[0].lower() + "".join(p.capitalize() for p in parts[1:])


def camel_case_to_all_upper(name: str) -> str:
    """
    Converts a string from camelCase to ALL_UPPER_CASE.
    """
    matches = _SPLIT_CAMEL_CASE_PATTERN.finditer(name)
    groups = [m.group(0).upper() for m in matches]
    return "_".join(groups)


def upper_camel_case_to_camel_case(name: str) -> str:
    if not name:
        return ""

    return name[0].lower() + name[1:]


def common_serialize(
    value: Any,
    value_type: Optional[Type],
    class_serializer: Serializer,
    serializers: Dict[Type, Serializer],
    user_serializers: Dict[Type, Callable[[Any], Any]],
) -> Any:
    # Find the type
    if value_type is None:
        value_type = type(value)

    # Is there a custom serializer for this class?
    try:
        serializer = user_serializers[value_type]
    except KeyError:
        pass
    else:
        return serializer(value)

    # Find a matching serializer
    key = get_origin(value_type)
    if key is None:
        key = value_type

    try:
        serializer = serializers[key]
    except KeyError:
        if inspect.isclass(value_type):
            serializer = class_serializer
        else:
            raise ValueError(f"Unsupported field of type {value_type}") from None

    # Define the recursion function
    def recur(value: Any, value_type: Type) -> Any:
        return common_serialize(
            value,
            value_type,
            class_serializer,
            serializers,
            user_serializers,
        )

    # Apply it
    return serializer(value, value_type, recur)


def common_deserialize(
    value: Any,
    value_type: Type,
    class_deserializer: Serializer,
    deserializers: Dict[Type, Deserializer],
    user_deserializers: Dict[Type, Callable[[Any], Any]],
) -> Any:
    # Is there a custom deserializer for this class?
    try:
        deserializer = user_deserializers[value_type]
    except KeyError:
        pass
    else:
        return deserializer(value)

    # Find a matching deserializer
    key = get_origin(value_type)
    if key is None:
        key = value_type

    try:
        deserializer = deserializers[key]
    except KeyError:
        if inspect.isclass(value_type):
            deserializer = class_deserializer
        else:
            raise ValueError(f"Unsupported field of type {value_type}") from None

    # Define the recursion function
    def recur(value: Any, value_type: Type) -> Any:
        return common_deserialize(
            value,
            value_type,
            class_deserializer,
            deserializers,
            user_deserializers,
        )

    # Apply it
    return deserializer(value, value_type, recur)


def all_subclasses(cls: Type, include_cls: bool) -> Iterable[Type]:
    """
    Yields all classes directly on indirectly inheriting from `cls`. Does not
    perform any sort of cycle checks.

    :param cls: The class to start from.
    :param include_cls: Whether to include `cls` itself in the results.
    """

    if include_cls:
        yield cls

    for subclass in cls.__subclasses__():
        yield from all_subclasses(subclass, include_cls=True)



from typing import Any, Callable, Dict, Optional, Type, TypeVar

from . import schema_mongodb, serde_bson, serde_json

T = TypeVar("T", bound="Serde")


class Serde:
    def as_bson(
        self,
        *,
        as_type: Optional[Type] = None,
        custom_serializers: Dict[Type, Callable[[Any], Any]] = {},
    ) -> Dict[str, Any]:
        """
        Serialize the entire instance to BSON, by applying the field serializer
        to each field. Field names are converted to camel case.
        """
        assert as_type is None or issubclass(self.__class__, as_type), as_type
        return serde_bson.as_bson(
            self,
            as_type=as_type,
            custom_serializers=custom_serializers,
        )

    def as_json(
        self,
        *,
        as_type: Optional[Type] = None,
        custom_serializers: Dict[Type, Callable[[Any], Any]] = {},
    ) -> Dict[str, Any]:
        """
        Serialize the entire instance to JSON, by applying the field serializer
        to each field. Field names are converted to camel case.
        """
        assert as_type is None or issubclass(self.__class__, as_type), as_type
        return serde_json.as_json(
            self,
            as_type=as_type,
            custom_serializers=custom_serializers,
        )

    @classmethod
    def from_bson(
        cls: Type[T],
        document: Dict[str, Any],
        custom_deserializers: Dict[Type, Callable[[Any], Any]] = {},
    ) -> T:
        """
        Deserialize an entire data class from BSON, by applying the field
        deserializer to each field. Field names are converted to camel case. The
        value may be modified by this function!
        """
        return serde_bson.from_bson(
            document,
            as_type=cls,
            custom_deserializers=custom_deserializers,
        )

    @classmethod
    def from_json(
        cls: Type[T],
        document: Dict[str, Any],
        custom_deserializers: Dict[Type, Callable[[Any], Any]] = {},
    ) -> T:
        """
        Deserialize an entire data class from JSON, by applying the field
        deserializer to each field. Field names are converted to camel case. The
        value may be modified by this function!
        """
        return serde_json.from_json(
            document,
            as_type=cls,
            custom_deserializers=custom_deserializers,
        )

    @classmethod
    def as_mongodb_schema(
        cls,
        custom_handlers: Dict[Type, Callable[[Any], Any]] = {},
    ) -> Any:
        return schema_mongodb.as_mongodb_schema(
            cls,
            custom_handlers=custom_handlers,
        )



import inspect
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, Literal, Type, Union, get_args, get_type_hints

from bson import ObjectId

from . import serde_class
from .common import *


def make_schema_bool_to_bool(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"type": "boolean"}


def make_schema_int_to_int(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"bsonType": ["int", "long"]}


def make_schema_float_to_float(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"bsonType": ["int", "long", "double"]}


def make_schema_bytes_to_bytes(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"bsonType": "binData"}


def make_schema_str_to_str(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"type": "string"}


def make_schema_datetime_to_datetime(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"bsonType": "date"}


def make_schema_list_to_list(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {
        "type": "array",
        "items": recur(None, get_args(value_type)[0]),
    }


def make_schema_dict_to_dict(
    none: None,
    value_type: Type,
    recur: Recur,
):
    subtypes = get_args(value_type)
    assert subtypes[0] is str, value_type

    return {
        "type": "object",
        "items": recur(None, subtypes[1]),
    }


def make_schema_object_id_to_object_id(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"bsonType": "objectId"}


def make_schema_literal_to_str(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {"type": "string"}


def make_schema_union(
    none: None,
    value_type: Type,
    recur: Recur,
):
    # Convert each subtype to a BSON schema
    sub_schemas = []
    for subtype in get_args(value_type):
        # Union is used by Python to represent "Optional"
        if subtype is type(None):
            sub_schemas.append({"type": "null"})
            continue

        sub_schemas.append(recur(None, subtype))

    # Prettify the result: instead of `{anyof {type ...} {type ...}}` just
    # create one `type`
    types = []
    bson_types = []
    others = []

    for schema in sub_schemas:
        if len(schema) == 1:

            # Standard Json Schema type
            try:
                type_field = schema["type"]
            except KeyError:
                pass
            else:
                if isinstance(type_field, list):
                    types.extend(type_field)
                else:
                    types.append(type_field)

                continue

            # BSON type
            try:
                type_field = schema["bsonType"]
            except KeyError:
                pass
            else:
                if isinstance(type_field, list):
                    bson_types.extend(type_field)
                else:
                    bson_types.append(type_field)

                continue

        # General case
        others.append(schema)

    # Create new, merged schemas
    sub_schemas = []

    if bson_types:
        sub_schemas.append({"bsonType": types + bson_types})
    elif types:
        sub_schemas.append({"type": types})

    sub_schemas.extend(others)

    if len(sub_schemas) == 1:
        return sub_schemas[0]

    return {"anyOf": sub_schemas}


def make_schema_any(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {}


def make_schema_tuple_to_list(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {
        "type": "array",
        "items": [recur(None, subtype) for subtype in get_args(value_type)],
    }


def make_schema_set_to_list(
    none: None,
    value_type: Type,
    recur: Recur,
):
    return {
        "type": "array",
        "items": recur(None, get_args(value_type)[0]),
    }


def create_class_schema_ignore_serialize_as_child(value_type: Type, recur: Recur):
    doc_field_names = []
    doc_properties = {}
    result = {
        "type": "object",
        "properties": doc_properties,
        "additionalProperties": False,
    }

    for field_py_name, field_type in get_type_hints(value_type).items():
        field_doc_name = all_lower_to_camel_case(field_py_name)

        if field_py_name == "id":
            field_doc_name = "_id"

        doc_field_names.append(field_doc_name)
        doc_properties[field_doc_name] = recur(None, field_type)

    # The `required` field may only be present if it contains at least one value
    if doc_field_names:
        result["required"] = doc_field_names

    return result


def make_schema_class(
    none: None,
    value_type: Type,
    recur: Recur,
):
    assert inspect.isclass(value_type), value_type

    # Case: The class has a custom schema method
    try:
        override_method = getattr(value_type, "as_mongodb_schema")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.as_mongodb_schema.__func__:
            return override_method()

    # Case: Enum
    if issubclass(value_type, Enum):
        return {
            "enum": [all_upper_to_camel_case(variant.name) for variant in value_type],
        }

    # Case: Class, and definitely not one of it's children
    if not should_serialize_as_child(value_type):
        return create_class_schema_ignore_serialize_as_child(value_type, recur)

    # Case: Class or one of its children

    # Create the schemas for all allowable classes
    sub_schemas = []
    for subtype in all_subclasses(value_type, True):
        schema = create_class_schema_ignore_serialize_as_child(subtype, recur)
        assert schema["type"] == "object", schema

        schema["properties"]["type"] = {
            "enum": [upper_camel_case_to_camel_case(subtype.__name__)]
        }

        required = schema.setdefault("required", [])
        required.insert(0, "type")

        sub_schemas.append(schema)

    # Create the final, combined schema
    if len(sub_schemas) == 1:
        return sub_schemas[0]
    else:
        return {"anyOf": sub_schemas}


SCHEMA_MAKERS: Dict[Type, Serializer] = {
    bool: make_schema_bool_to_bool,
    int: make_schema_int_to_int,
    float: make_schema_float_to_float,
    bytes: make_schema_bytes_to_bytes,
    str: make_schema_str_to_str,
    datetime: make_schema_datetime_to_datetime,
    list: make_schema_list_to_list,
    dict: make_schema_dict_to_dict,
    Union: make_schema_union,
    Any: make_schema_any,
    ObjectId: make_schema_object_id_to_object_id,
    Literal: make_schema_literal_to_str,
    tuple: make_schema_tuple_to_list,
    set: make_schema_set_to_list,
}


def as_mongodb_schema(
    value_type: Type,
    *,
    custom_handlers: Dict[Type, Callable[[Any], Any]] = {},
) -> Any:
    return common_serialize(
        None,
        value_type,
        make_schema_class,
        SCHEMA_MAKERS,
        custom_handlers,
    )



from datetime import datetime, timezone
from typing import Any, Dict, Optional

from bson import ObjectId

from . import serde_class, serde_json
from .common import *

__all__ = [
    "as_bson",
    "from_bson",
]


def serialize_bytes_to_bytes(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bytes:
    assert isinstance(value, bytes), value
    return value


def serialize_datetime_to_datetime(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> datetime:
    assert isinstance(value, datetime), value
    assert (
        value.tzinfo is not None
    ), f"Encountered datetime without a timezone. Please always set timezones, or expect hard to find bugs."

    return value


def serialize_object_id_to_object_id(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> ObjectId:
    assert isinstance(value, ObjectId), value
    return value


def serialize_class(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Dict[str, Any]:
    result = serde_json.serialize_class(value, value_type, recur)

    # Case: The class has a custom serialization method
    try:
        override_method = getattr(value, "as_bson")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.as_bson:
            return override_method()

    # Map "id" to "_id", as is done in MongoDB.
    if isinstance(result, dict) and "_id" not in result:
        try:
            result["_id"] = result.pop("id")
        except KeyError:
            pass

    return result


BSON_SERIALIZERS: Dict[Type, Serializer] = serde_json.JSON_SERIALIZERS.copy()
BSON_SERIALIZERS.update(
    {
        bytes: serialize_bytes_to_bytes,
        datetime: serialize_datetime_to_datetime,
        ObjectId: serialize_object_id_to_object_id,
    }
)


def as_bson(
    value: Any,
    *,
    as_type: Optional[Type] = None,
    custom_serializers: Dict[Type, Callable[[Any], Any]] = {},
) -> Any:
    return common_serialize(
        value,
        type(value) if as_type is None else as_type,
        serialize_class,
        BSON_SERIALIZERS,
        custom_serializers,
    )


def deserialize_bytes_from_bytes(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bytes:
    if not isinstance(value, bytes):
        raise SerdeError(f"Expected bytes, got {value}")

    return value


def deserialize_datetime_from_datetime(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> datetime:
    if not isinstance(value, datetime):
        raise SerdeError(f"Expected datetime, got {value}")

    # BSON doesn't support timezones, and MongoDB convention dictates UTC to be
    # assumed. Impute that.
    if value.tzinfo is None:
        value = value.replace(tzinfo=timezone.utc)

    return value


def deserialize_object_id_from_object_id(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> ObjectId:
    if not isinstance(value, ObjectId):
        raise SerdeError(f"Expected ObjectId, got {value}")

    return value


def deserialize_class(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Dict[str, Any]:
    # Case: The class has a custom deserialization method
    try:
        override_method = getattr(value_type, "from_bson")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.from_bson.__func__:
            return override_method(value)

    # Map "_id" to "id"
    if isinstance(value, dict) and "id" not in value:
        try:
            value["id"] = value.pop("_id")
        except KeyError:
            pass

    # Then hand off to the JSON deserializer.
    return serde_json.deserialize_class(value, value_type, recur)


BSON_DESERIALIZERS: Dict[Type, Deserializer] = serde_json.JSON_DESERIALIZERS.copy()
BSON_DESERIALIZERS.update(
    {
        bytes: deserialize_bytes_from_bytes,
        datetime: deserialize_datetime_from_datetime,
        ObjectId: deserialize_object_id_from_object_id,
    }
)


def from_bson(
    value: Any,
    as_type: Type,
    *,
    custom_deserializers: Dict[Type, Callable[[Any], Any]] = {},
) -> Any:
    return common_deserialize(
        value,
        as_type,
        deserialize_class,
        BSON_DESERIALIZERS,
        custom_deserializers,
    )



import base64
import inspect
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Type, Union, get_args, get_type_hints

import dateutil.parser
from bson import ObjectId

from . import serde_class
from .common import *

__all__ = [
    "as_json",
    "from_json",
]


def serialize_bool_to_bool(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bool:
    assert isinstance(value, bool), value
    return value


def serialize_int_to_int(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> int:
    assert isinstance(value, int), value
    return value


def serialize_float_to_float(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> float:
    assert isinstance(value, (int, float)), value
    return float(value)


def serialize_bytes_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, bytes), value
    return base64.b64encode(value).decode("utf-8")


def serialize_str_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, str), value
    return value


def serialize_datetime_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, datetime), value
    assert (
        value.tzinfo is not None
    ), f"Encountered datetime without a timezone. Please always set timezones, or expect hard to find bugs."
    return value.isoformat()


def serialize_list_to_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> List[Any]:
    assert isinstance(value, list), value
    return [recur(v, get_args(value_type)[0]) for v in value]


def serialize_dict_to_dict(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Dict[Any, Any]:
    subtypes = get_args(value_type)
    assert isinstance(value, dict), value
    return {recur(k, subtypes[0]): recur(v, subtypes[1]) for k, v in value.items()}


def serialize_object_id_to_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    assert isinstance(value, ObjectId), value
    return str(value)


def serialize_optional(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    if value is None:
        return None

    return recur(value, get_args(value_type)[0])


def serialize_any(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    return value


def serialize_literal_as_is(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    assert isinstance(value, str), value
    return value


def serialize_tuple_as_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> List[Any]:
    assert isinstance(value, tuple), value
    subtypes = get_args(value_type)
    assert len(subtypes) == len(value), (subtypes, value)

    return [recur(v, subtype) for v, subtype in zip(value, subtypes)]


def serialize_set_as_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> List[Any]:
    assert isinstance(value, set), value
    subtype = get_args(value_type)[0]
    return [recur(v, subtype) for v in value]


def serialize_class(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    assert inspect.isclass(value_type), value_type

    # Case: The class has a custom serialization method
    try:
        override_method = getattr(value, "as_json")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.as_json:
            return override_method()

    # Case: Enum
    if issubclass(value_type, Enum):
        assert isinstance(value, value_type), value
        return all_upper_to_camel_case(value.name)

    # Case: Anything else
    # Make sure to serialize as the correct class
    if should_serialize_as_child(value_type):
        assert issubclass(type(value), value_type), (type(value), value_type)
        value_type = type(value)

    result = {}
    for field_py_name, field_type in get_type_hints(value_type).items():
        field_doc_name = all_lower_to_camel_case(field_py_name)
        result[field_doc_name] = recur(getattr(value, field_py_name), field_type)

    # Add a type tag?
    if should_serialize_as_child(value_type):
        result["type"] = upper_camel_case_to_camel_case(value.__class__.__name__)

    return result


JSON_SERIALIZERS: Dict[Type, Serializer] = {
    bool: serialize_bool_to_bool,
    int: serialize_int_to_int,
    float: serialize_float_to_float,
    bytes: serialize_bytes_to_str,
    str: serialize_str_to_str,
    datetime: serialize_datetime_to_str,
    list: serialize_list_to_list,
    dict: serialize_dict_to_dict,
    Union: serialize_optional,
    Any: serialize_any,
    ObjectId: serialize_object_id_to_str,
    Literal: serialize_literal_as_is,
    tuple: serialize_tuple_as_list,
    set: serialize_set_as_list,
}


def as_json(
    value: Any,
    *,
    as_type: Optional[Type] = None,
    custom_serializers: Dict[Type, Callable[[Any], Any]] = {},
) -> Any:
    return common_serialize(
        value,
        type(value) if as_type is None else as_type,
        serialize_class,
        JSON_SERIALIZERS,
        custom_serializers,
    )


def deserialize_bool_from_bool(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bool:
    if not isinstance(value, bool):
        raise SerdeError(f"Expected bool, got {value}")

    return value


def deserialize_int_from_int(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> int:
    if not isinstance(value, (float, int)) or int(value) != value:
        raise SerdeError(f"Expected int, got {value}")

    return int(value)


def deserialize_float_from_float(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> float:
    if not isinstance(value, (int, float)):
        raise SerdeError(f"Expected float, got {value}")

    return float(value)


def deserialize_bytes_from_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> bytes:
    if not isinstance(value, str):
        raise SerdeError(f"Expected bytes encoded as string, got {value}")

    try:
        return base64.b64decode(value.encode("utf-8"))
    except ValueError:
        raise SerdeError("Received inavlid base64 encoded string.")


def deserialize_str_from_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> str:
    if not isinstance(value, str):
        raise SerdeError(f"Expected string, got {value}")

    return value


def deserialize_datetime_from_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> datetime:
    if not isinstance(value, str):
        raise SerdeError(f"Expected date/time string, got {value}")

    try:
        result = dateutil.parser.isoparse(value)
    except ValueError:
        raise SerdeError(f"Expected date/time, got {value}") from None

    if result.tzinfo is None:
        raise SerdeError(f"The date/time value {value} is missing a timezone.")

    return result


def deserialize_list_from_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> List[Any]:
    if not isinstance(value, list):
        raise SerdeError(f"Expected list, got {value}")

    subtype = get_args(value_type)[0]

    return [recur(v, subtype) for v in value]


def deserialize_dict_from_dict(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Dict[Any, Any]:
    if not isinstance(value, dict):
        raise SerdeError(f"Expected dict, got {value}")

    subtypes = get_args(value_type)

    return {recur(k, subtypes[0]): recur(v, subtypes[1]) for k, v in value.items()}


def deserialize_object_id_from_str(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> ObjectId:
    if not isinstance(value, str):
        raise SerdeError(f"Expected ObjectId string, got {value}")

    try:
        result = ObjectId(value)
    except ValueError:
        raise SerdeError(f"Expected ObjectId string, got {value}") from None

    return result


def deserialize_optional(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    if value is None:
        return None

    return recur(value, get_args(value_type)[0])


def deserialize_any(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    return value


def deserialize_literal_as_is(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    options = get_args(value_type)
    if value not in options:
        raise SerdeError(f"Expected {value_type}, got {value}")

    return value


def deserialize_tuple_from_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    if not isinstance(value, list):
        raise SerdeError(f"Expected list, got {value}")

    subtypes = get_args(value_type)

    if len(value) != len(subtypes):
        raise SerdeError(
            f"Expected list of size {len(subtypes)}, but received one of size {len(value)}"
        )

    return tuple(recur(v, subtype) for v, subtype in zip(value, subtypes))


def deserialize_set_from_list(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    if not isinstance(value, list):
        raise SerdeError(f"Expected list, got {value}")

    subtype = get_args(value_type)[0]

    return set(recur(v, subtype) for v in value)


def deserialize_class(
    value: Any,
    value_type: Type,
    recur: Recur,
) -> Any:
    assert inspect.isclass(value_type), value_type

    # Case: The class has a custom deserialization method
    try:
        override_method = getattr(value_type, "from_json")
    except AttributeError:
        pass
    else:
        if override_method.__func__ is not serde_class.Serde.from_json.__func__:
            return override_method(value)

    # Case: Enum
    if issubclass(value_type, Enum):
        if not isinstance(value, str):
            raise SerdeError(f"Expected enumeration string, got {value}")

        try:
            py_name = camel_case_to_all_upper(value)  # ValueError if not camel case
            return value_type[py_name]  # ValueError if not in enum
        except KeyError:
            raise SerdeError(f"Invalid enumeration value {value}") from None

    # Should this class be deserialized as a child? If so, find the actual
    # matching child class.
    if should_serialize_as_child(value_type):
        try:
            type_tag = value.pop("type")
        except KeyError:
            raise SerdeError(f'Object is missing the "type" field') from None

        candidate_classes = all_subclasses(value_type, True)

        for actual_type in candidate_classes:
            if upper_camel_case_to_camel_case(actual_type.__name__) == type_tag:
                break
        else:
            raise SerdeError(f'Object has an invalid type tag "{type_tag}"') from None

    else:
        actual_type = value_type

    del value_type

    # Case: Anything else
    if not isinstance(value, dict):
        raise SerdeError(f"Expected Object, got {value}")

    # Deserialize each field
    result = object.__new__(actual_type)
    for field_name, field_type in get_type_hints(actual_type).items():
        doc_name = all_lower_to_camel_case(field_name)

        try:
            field_value = recur(value.pop(doc_name), field_type)
        except KeyError:
            raise SerdeError(f'Object is missing the "{doc_name}" field') from None

        result.__dict__[field_name] = field_value  # type: ignore

    # Check for any superfluous fields
    if len(value) > 0:
        raise SerdeError(
            f"Superfluous object fields {', '.join(map(str, value.keys()))}"
        )

    return result


JSON_DESERIALIZERS: Dict[Type, Deserializer] = {
    bool: deserialize_bool_from_bool,
    int: deserialize_int_from_int,
    float: deserialize_float_from_float,
    bytes: deserialize_bytes_from_str,
    str: deserialize_str_from_str,
    datetime: deserialize_datetime_from_str,
    list: deserialize_list_from_list,
    dict: deserialize_dict_from_dict,
    Union: deserialize_optional,
    Any: deserialize_any,
    ObjectId: deserialize_object_id_from_str,
    Literal: deserialize_literal_as_is,
    tuple: deserialize_tuple_from_list,
    set: deserialize_set_from_list,
}


def from_json(
    value: Any,
    as_type: Type,
    *,
    custom_deserializers: Dict[Type, Callable[[Any], Any]] = {},
) -> Any:
    return common_deserialize(
        value,
        as_type,
        deserialize_class,
        JSON_DESERIALIZERS,
        custom_deserializers,
    )
